/*!\file Node.c
 * \brief: implementation of the Node object
 */

/*Include files: {{{1*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#error "Cannot compile with HAVE_CONFIG_H symbol! run configure first!"
#endif

#include "stdio.h"
#include <string.h>
#include "./objects.h"
#include "../Container/Container.h"
#include "../EnumDefinitions/EnumDefinitions.h"
#include "../shared/shared.h"
#include "../include/include.h"
#include "../modules/modules.h"
/*}}}*/

/*Node constructors and destructors:*/
/*FUNCTION Node::Node() default constructor {{{1*/
Node::Node(){
	this->inputs=NULL;
	this->hvertex=NULL;
	return;
}
/*}}}*/
/*FUNCTION Node::Node(int id, int vertex_id, int numdofs, NodeProperties*) {{{1*/
Node::Node(int node_id,int node_sid, int node_vertex_id, int node_numdofs):
	indexing(node_numdofs){

	this->id=node_id;
	this->sid=node_sid;
	this->inputs=new Inputs();
	this->hvertex=new Hook(&node_vertex_id,1);

}
/*}}}*/
/*FUNCTION Node::Node(int id, DofIndexing* indexing, Hook* vertex, Inputs* inputs,int analysis_type){{{1*/
Node::Node(int node_id,int node_sid, DofIndexing* node_indexing, Hook* node_vertex,Inputs* node_inputs,int analysis_type):
	    indexing(node_indexing),
		hvertex(node_vertex)
{

	/*all the initialization has been done by the initializer, just fill in the id: */
	this->id=node_id;
	this->sid=node_sid;
	this->analysis_type=analysis_type;

	if(node_inputs){
		this->inputs=(Inputs*)node_inputs->Copy();
	}
	else{
		this->inputs=new Inputs();
	}
}
/*}}}*/
/*FUNCTION Node::Node(int node_id,int vertex_id,int io_index, IoModel* iomodel,int analysis_type) {{{1*/
Node::Node(int node_id,int node_sid,int vertex_id,int io_index, IoModel* iomodel,int analysis_type){

	/*Intermediary*/
	int k;
	int numdofs;

	/*id: */
	this->id=node_id; 
	this->sid=node_sid; 
	this->analysis_type=analysis_type;

	/*indexing:*/
	DistributeNumDofs(&numdofs,analysis_type,iomodel->vertices_type+2*io_index); //number of dofs per node
	this->indexing.Init(numdofs);

	/*Hooks*/
	this->hvertex=new Hook(&vertex_id,1); //node id is the same as the vertex id, continuous galerkin!

	//intialize inputs, and add as many inputs per element as requested: 
	this->inputs=new Inputs();
	if (iomodel->gridonbed)      this->inputs->AddInput(new BoolInput(NodeOnBedEnum,(IssmBool)iomodel->gridonbed[io_index]));
	if (iomodel->gridonsurface)  this->inputs->AddInput(new BoolInput(NodeOnSurfaceEnum,(IssmBool)iomodel->gridonsurface[io_index]));
	if (iomodel->gridoniceshelf) this->inputs->AddInput(new BoolInput(NodeOnIceShelfEnum,(IssmBool)iomodel->gridoniceshelf[io_index]));
	if (iomodel->gridonicesheet) this->inputs->AddInput(new BoolInput(NodeOnIceSheetEnum,(IssmBool)iomodel->gridonicesheet[io_index]));
	if (iomodel->numbernodetoelementconnectivity) this->inputs->AddInput(new IntInput(NumberNodeToElementConnectivityEnum,iomodel->numbernodetoelementconnectivity[io_index]));

	/*set single point constraints: */

	/*Diagnostic Horiz*/
	if (analysis_type==DiagnosticHorizAnalysisEnum){
		if (iomodel->dim==3){
			/*We have a  3d mesh, we may have collapsed elements, hence dead grids. Freeze them out: */
			if (!iomodel->deadgrids) ISSMERROR("iomodel->deadgrids is NULL");
			if (iomodel->deadgrids[io_index]){
				for(k=1;k<=numdofs;k++){
					this->FreezeDof(k);
				}
			}
		}
		/*spc all nodes on hutter*/
		if (!iomodel->gridonhutter) ISSMERROR("iomodel->gridonhutter is NULL");
		if (iomodel->gridonhutter[io_index]){
			for(k=1;k<=numdofs;k++){
				this->FreezeDof(k);
			}
		}
	}

	/*Diagnostic Stokes*/
	if (analysis_type==DiagnosticStokesAnalysisEnum){
		/*On a 3d mesh, in stokes formualtions, only stokes grids are free, the others are frozen: */
		if (!iomodel->borderstokes) ISSMERROR("iomodel->borderstokes is NULL");
		if (iomodel->borderstokes[io_index]){
			//freeze everything except pressure
			this->FreezeDof(1);
			this->FreezeDof(2);
			this->FreezeDof(3);
		}
		else if (iomodel->gridonstokes[io_index]==0){
			for(k=1;k<=numdofs;k++){
				this->FreezeDof(k);
			}
		}
	}

	/*Diagnostic Hutter*/
	if (analysis_type==DiagnosticHutterAnalysisEnum){
		/*Spc all nodes that are not Hutter*/
		if (!iomodel->gridonhutter) ISSMERROR("iomodel->gridonhutter is NULL");
		if (!iomodel->gridonhutter[io_index]){
			for(k=1;k<=numdofs;k++){
				this->FreezeDof(k);
			}
		}
	}

	/*Prognostic/ Melting/ Slopecompute/ Balancedvelocities/ Balancedthickness*/
	if (
				analysis_type==PrognosticAnalysisEnum || 
				analysis_type==MeltingAnalysisEnum || 
				analysis_type==BedSlopeAnalysisEnum || 
				analysis_type==SurfaceSlopeAnalysisEnum || 
				analysis_type==BalancedvelocitiesAnalysisEnum || 
				analysis_type==BalancedthicknessAnalysisEnum
				){
		if (iomodel->dim==3){
			/*On a 3d mesh, we may have collapsed elements, hence dead grids. Freeze them out: */
			if (!iomodel->gridonbed) ISSMERROR("iomodel->gridonbed is NULL");
			if (!iomodel->gridonbed[io_index]){
				for(k=1;k<=numdofs;k++){
					this->FreezeDof(k);
				}
			}
		}
	}

}
/*}}}*/
/*FUNCTION Node::~Node(){{{1*/
Node::~Node(){
	delete inputs;
	delete hvertex;
	return;
}
/*}}}*/

/*Object virtual functions definitions:*/
/*FUNCTION Node::Echo{{{1*/
void Node::Echo(void){

	printf("Node:\n");
	printf("   id: %i\n",id);
	printf("   sid: %i\n",sid);
	printf("   analysis_type: %s\n",EnumToString(analysis_type));
	indexing.Echo();
	printf("   hvertex:     not displayed\n");
	printf("   inputs:      %p\n",inputs);


}
/*}}}*/
/*FUNCTION Node::DeepEcho{{{1*/
void Node::DeepEcho(void){

	printf("Node:\n");
	printf("   id: %i\n",id);
	printf("   sid: %i\n",sid);
	printf("   analysis_type: %s\n",EnumToString(analysis_type));
	indexing.DeepEcho();
	printf("Vertex:\n");
	hvertex->DeepEcho();
	printf("   inputs\n");


}
/*}}}*/
/*FUNCTION Node::Id{{{1*/
int    Node::Id(void){ return id; }
/*}}}*/
/*FUNCTION Node::MyRank{{{1*/
int    Node::MyRank(void){ 
	extern int my_rank;

	return my_rank; 
}
/*}}}*/
/*FUNCTION Node::Marshall{{{1*/
void  Node::Marshall(char** pmarshalled_dataset){

	char* marshalled_dataset=NULL;
	int   enum_type=0;
	char* marshalled_inputs=NULL;
	int   marshalled_inputs_size;

	/*recover marshalled_dataset: */
	marshalled_dataset=*pmarshalled_dataset;

	/*get enum type of Node: */
	enum_type=NodeEnum;
	
	/*marshall enum: */
	memcpy(marshalled_dataset,&enum_type,sizeof(enum_type));marshalled_dataset+=sizeof(enum_type);
	
	/*marshall Node data: */
	memcpy(marshalled_dataset,&id,sizeof(id));marshalled_dataset+=sizeof(id);
	memcpy(marshalled_dataset,&sid,sizeof(sid));marshalled_dataset+=sizeof(sid);
	memcpy(marshalled_dataset,&analysis_type,sizeof(analysis_type));marshalled_dataset+=sizeof(analysis_type);
	
	/*marshall objects: */
	indexing.Marshall(&marshalled_dataset);
	hvertex->Marshall(&marshalled_dataset);

	/*Marshall inputs: */
	marshalled_inputs_size=inputs->MarshallSize();
	marshalled_inputs=inputs->Marshall();
	memcpy(marshalled_dataset,marshalled_inputs,marshalled_inputs_size*sizeof(char));
	marshalled_dataset+=marshalled_inputs_size;

	/*Free ressources:*/
	xfree((void**)&marshalled_inputs);

	*pmarshalled_dataset=marshalled_dataset;
	return;
}
/*}}}*/
/*FUNCTION Node::MarshallSize{{{1*/
int   Node::MarshallSize(){

	return sizeof(id)+
		sizeof(sid)+
		indexing.MarshallSize()+
		hvertex->MarshallSize()+
		inputs->MarshallSize()+
		sizeof(analysis_type)+
		sizeof(int); //sizeof(int) for enum type
}
/*}}}*/
/*FUNCTION Node::Demarshall{{{1*/
void  Node::Demarshall(char** pmarshalled_dataset){

	char* marshalled_dataset=NULL;

	/*recover marshalled_dataset: */
	marshalled_dataset=*pmarshalled_dataset;

	/*this time, no need to get enum type, the pointer directly points to the beginning of the 
	 *object data (thanks to DataSet::Demarshall):*/

	memcpy(&id,marshalled_dataset,sizeof(id));marshalled_dataset+=sizeof(id);
	memcpy(&sid,marshalled_dataset,sizeof(sid));marshalled_dataset+=sizeof(sid);
	memcpy(&analysis_type,marshalled_dataset,sizeof(analysis_type));marshalled_dataset+=sizeof(analysis_type);
	
	/*demarshall objects: */
	indexing.Demarshall(&marshalled_dataset);
	hvertex=new Hook(); hvertex->Demarshall(&marshalled_dataset);

	/*demarshall inputs: */
	inputs=(Inputs*)DataSetDemarshallRaw(&marshalled_dataset); 

	/*return: */
	*pmarshalled_dataset=marshalled_dataset;
	return;
}
/*}}}*/
/*FUNCTION Node::Enum{{{1*/
int Node::Enum(void){

	return NodeEnum;

}
/*}}}*/
/*FUNCTION Node::copy {{{1*/
Object* Node::copy() {
		
	return new Node(this->id,this->sid,&this->indexing, this->hvertex,this->inputs,this->analysis_type);

}

/*}}}*/

/*Node management:*/
/*FUNCTION Node::Configure {{{1*/
void  Node::Configure(DataSet* nodesin,Vertices* verticesin){

	int i;

	/*Take care of hooking up all objects for this element, ie links the objects in the hooks to their respective 
	 * datasets, using internal ids and offsets hidden in hooks: */
	hvertex->configure(verticesin);

}
/*FUNCTION Node::SetCurrentConfiguration {{{1*/
void  Node::SetCurrentConfiguration(DataSet* nodesin,Vertices* verticesin){

}
/*FUNCTION Node::GetDof {{{1*/
int   Node::GetDof(int dofindex){

	return indexing.doflist[dofindex];

}
/*}}}*/
/*FUNCTION Node::GetDofList1{{{1*/
int  Node::GetDofList1(void){

	Vertex* vertex=NULL;

	vertex=(Vertex*)this->hvertex->delivers();

	return vertex->dof;
}
/*}}}*/
/*FUNCTION Node::GetDofList{{{1*/
void  Node::GetDofList(int* outdoflist){
	int i;
	for(i=0;i<this->indexing.numberofdofs;i++){
		outdoflist[i]=indexing.doflist[i];
	}
}
/*}}}*/
/*FUNCTION Node::Sid{{{1*/
int    Node::Sid(void){ return sid; }
/*}}}*/
/*FUNCTION Node::GetVertexId {{{1*/
int   Node::GetVertexId(void){

	Vertex*  vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	return vertex->id;
}
/*}}}*/
/*FUNCTION Node::GetVertexDof {{{1*/
int   Node::GetVertexDof(void){

	Vertex*  vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	return vertex->dof;
}
/*}}}*/
/*FUNCTION Node::SetVertexDof {{{1*/
void   Node::SetVertexDof(int in_dof){

	Vertex*  vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	vertex->dof=in_dof;

}
/*}}}*/
/*FUNCTION Node::InAnalysis{{{1*/
bool Node::InAnalysis(int in_analysis_type){
	if (in_analysis_type==this->analysis_type) return true;
	else return false;
}
/*}}}*/

/*Node numerics:*/
/*FUNCTION Node::ApplyConstraints{{{1*/
void  Node::ApplyConstraint(Vec yg,int dof,double value){

	int index;

	/*First, dof should be added in the s set, describing which 
	 * dofs are constrained to a certain value (dirichlet boundary condition*/

	DofInSSet(dof-1);

	/*Second, we should add value into yg, at dof corresponding to doflist[dof], unless
	 *  we are a clone!*/

	if(!indexing.clone){

		index=indexing.doflist[dof-1]; //matlab indexing

		VecSetValues(yg,1,&index,&value,INSERT_VALUES);

	}

}
/*}}}*/
/*FUNCTION Node::CreateVecSets {{{1*/
void  Node::CreateVecSets(Vec pv_g,Vec pv_f,Vec pv_s){

	double gvalue=1.0; //all nodes are in the g set;
	double value;

	int i;

	for(i=0;i<this->indexing.numberofdofs;i++){

		/*g set: */
		VecSetValues(pv_g,1,&indexing.doflist[i],&gvalue,INSERT_VALUES);
		
		/*f set: */
		value=(double)this->indexing.f_set[i];
		VecSetValues(pv_f,1,&indexing.doflist[i],&value,INSERT_VALUES);

		/*s set: */
		value=(double)this->indexing.s_set[i];
		VecSetValues(pv_s,1,&indexing.doflist[i],&value,INSERT_VALUES);

	}


}
/*}}}*/
/*FUNCTION Node::DofInSSet {{{1*/
void  Node::DofInSSet(int dof){

	/*Put dof for this node into the s set (ie, this dof will be constrained 
	 * to a fixed value during computations. */

	this->indexing.f_set[dof]=0; //n splits into f (for which we solve) and s (single point constraints)
	this->indexing.s_set[dof]=1;
}
/*}}}*/
/*FUNCTION Node::FreezeDof{{{1*/
void  Node::FreezeDof(int dof){
	
	DofInSSet(dof-1); //with 0 displacement for this dof.

}
/*}}}*/
/*FUNCTION Node::GetConnectivity {{{1*/
int Node::GetConnectivity(){
	int connectivity;

	/*recover parameters: */
	inputs->GetParameterValue(&connectivity,NumberNodeToElementConnectivityEnum);

	return connectivity;
}
/*}}}*/
/*FUNCTION Node::GetNumberOfDofs{{{1*/
int   Node::GetNumberOfDofs(){
	
	return this->indexing.numberofdofs;

}
/*}}}*/
/*FUNCTION Node::GetSigma {{{1*/
double Node::GetSigma(){
	Vertex* vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	return vertex->sigma;
}
/*}}}*/
/*FUNCTION Node::GetX {{{1*/
double Node::GetX(){
	Vertex* vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	return vertex->x;
}
/*}}}*/
/*FUNCTION Node::GetY {{{1*/
double Node::GetY(){
	Vertex* vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	return vertex->y;
}
/*}}}*/
/*FUNCTION Node::GetZ {{{1*/
double Node::GetZ(){
	Vertex* vertex=NULL;

	vertex=(Vertex*)hvertex->delivers();
	return vertex->z;
}
/*}}}*/
/*FUNCTION Node::IsClone {{{1*/
int   Node::IsClone(){
	
	return indexing.clone;

}
/*}}}*/
/*FUNCTION Node::IsOnBed {{{1*/
int   Node::IsOnBed(){

	bool onbed;

	/*recover parameters: */
	inputs->GetParameterValue(&onbed,NodeOnBedEnum);

	return onbed;
}
/*}}}*/
/*FUNCTION Node::IsOnSheet {{{1*/
int   Node::IsOnSheet(){

	bool onsheet;

	/*recover parameters: */
	inputs->GetParameterValue(&onsheet,NodeOnIceSheetEnum);

	return onsheet;
}		
/*}}}*/
/*FUNCTION Node::IsOnShelf {{{1*/
int   Node::IsOnShelf(){
	
	bool onshelf;

	/*recover parameters: */
	inputs->GetParameterValue(&onshelf,NodeOnIceShelfEnum);

	return onshelf;
}
/*}}}*/
/*FUNCTION Node::IsOnSurface {{{1*/
int   Node::IsOnSurface(){

	bool onsurface;

	/*recover parameters: */
	inputs->GetParameterValue(&onsurface,NodeOnSurfaceEnum);

	return onsurface;
}
/*}}}*/
/*FUNCTION Node::InputUpdateFromVector(double* vector, int name, int type){{{1*/
void  Node::InputUpdateFromVector(double* vector, int name, int type){

	/*Nothing updated yet*/
}
/*}}}*/
/*FUNCTION Node::InputUpdateFromVector(int* vector, int name, int type){{{1*/
void  Node::InputUpdateFromVector(int* vector, int name, int type){

	/*Nothing updated yet*/
}
/*}}}*/
/*FUNCTION Node::InputUpdateFromVector(bool* vector, int name, int type){{{1*/
void  Node::InputUpdateFromVector(bool* vector, int name, int type){

	/*Nothing updated yet*/
}
/*}}}*/
/*FUNCTION Node::InputUpdateFromConstant(double constant, int name){{{1*/
void  Node::InputUpdateFromConstant(double constant, int name){

	/*Nothing updated yet*/
}
/*}}}*/
/*FUNCTION Node::InputUpdateFromConstant(int constant, int name){{{1*/
void  Node::InputUpdateFromConstant(int constant, int name){

	/*Nothing updated yet*/
}
/*}}}*/
/*FUNCTION Node::InputUpdateFromConstant(bool constant, int name){{{1*/
void  Node::InputUpdateFromConstant(bool constant, int name){

	/*Nothing updated yet*/
}
/*}}}*/

/* DofObject routines:*/
/*FUNCTION Node::DistributeDofs{{{1*/
void  Node::DistributeDofs(int* pdofcount){

	int i;
	extern int my_rank;
	int dofcount;

	dofcount=*pdofcount;
	
	if(indexing.clone){
		/*This node is a clone! Don't distribute dofs, it will get them from another cpu!*/
		return;
	}

	/*This node should distribute dofs, go ahead: */
	for(i=0;i<this->indexing.numberofdofs;i++){
		indexing.doflist[i]=dofcount+i;
	}
	dofcount+=this->indexing.numberofdofs;

	/*Assign output pointers: */
	*pdofcount=dofcount;

}
/*}}}*/
/*FUNCTION Node::OffsetDofs{{{1*/
void  Node::OffsetDofs(int dofcount){
	
	int i;
	extern int my_rank;
	
	if(indexing.clone){
		/*This node is a clone, don't offset the dofs!: */
		return;
	}

	/*This node should offset the dofs, go ahead: */
	for(i=0;i<this->indexing.numberofdofs;i++){
		indexing.doflist[i]+=dofcount;
	}
}
/*}}}*/
/*FUNCTION Node::ShowTrueDofs{{{1*/
void  Node::ShowTrueDofs(int* truedofs, int ncols){

	int j;
	extern int my_rank;
	
	/*Are we a clone? : */
	if(indexing.clone)return;

	/*Ok, we are not a clone, just plug our dofs into truedofs: */
	for(j=0;j<this->indexing.numberofdofs;j++){
		*(truedofs+ncols*sid+j)=indexing.doflist[j];
	}
}
/*}}}*/
/*FUNCTION Node::UpdateCloneDofs{{{1*/
void  Node::UpdateCloneDofs(int* alltruedofs,int ncols){

	int j;
	extern int my_rank;
	
	/*If we are not a clone, don't update, we already have dofs!: */
	if(indexing.clone==0)return;


	/*Ok, we are a clone node, but we did not create the dofs for this node.
	 *      * Therefore, our doflist is garbage right now. Go pick it up in the alltruedofs: */
	for(j=0;j<this->indexing.numberofdofs;j++){
		indexing.doflist[j]=*(alltruedofs+ncols*sid+j);
	}

}
/*}}}*/
/*FUNCTION Node::SetClone {{{1*/
void  Node::SetClone(int* minranks){

	extern int my_rank;

	if (minranks[sid]==my_rank){
		indexing.clone=0;
	}
	else{
		/*!there is a cpu with lower rank that has the same node, 
		therefore, I am a clone*/
		indexing.clone=1; 	
	}

}
/*}}}*/
/*FUNCTION Node::CreatePartition{{{1*/
void  Node::CreatePartition(Vec partition){ 

	ISSMERROR(" not supported yet!");
	return;
}
/*}}}*/
